@@ -250,10 +250,8 @@ void widget_draw(widget *wid, cairo_t *d) {
     cairo_restore(d);
 
     if (left != 0 || top != 0 || right != 0 || bottom != 0) {
-      // NOTE: Cairo group push/pop has same effect as cairo_save/cairo_restore,
-      // thus no need for these.
-      cairo_push_group(d);
-      cairo_set_operator(d, CAIRO_OPERATOR_ADD);
+      cairo_save(d);
+      //      cairo_set_operator(d, CAIRO_OPERATOR_ADD);
       cairo_translate(d, wid->x, wid->y);
       cairo_new_path(d);
       rofi_theme_get_color(wid, "border-color", d);
@@ -409,8 +407,7 @@ void widget_draw(widget *wid, cairo_t *d) {
 
         cairo_fill(d);
       }
-      cairo_pop_group_to_source(d);
-      cairo_paint(d);
+      cairo_restore(d);
     }
   }
 }
