@@ -3,7 +3,7 @@
  *
  * MIT/X11 License
  * Copyright © 2012 Sean Pringle <sean.pringle@gmail.com>
- * Copyright © 2013-2021 Qball Cow <qball@gmpclient.org>
+ * Copyright © 2013-2023 Qball Cow <qball@gmpclient.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining
  * a copy of this software and associated documentation files (the
@@ -30,6 +30,10 @@
 #define G_LOG_DOMAIN "X11Helper"
 
 #include "config.h"
+#ifdef XCB_IMDKIT
+#include <xcb-imdkit/encoding.h>
+#include <xcb/xcb_keysyms.h>
+#endif
 #include <cairo-xcb.h>
 #include <cairo.h>
 #include <glib.h>
@@ -62,6 +66,10 @@
 #include "xcb-internal.h"
 #include "xcb.h"
 #include <libsn/sn.h>
+#include <stdbool.h>
+
+#include "mode.h"
+#include "modes/window.h"
 
 #include <rofi.h>
 
@@ -81,10 +89,15 @@ WindowManagerQuirk current_window_manager = WM_EWHM;
  */
 struct _xcb_stuff xcb_int = {.connection = NULL,
                              .screen = NULL,
+#ifdef XCB_IMDKIT
+                             .im = NULL,
+                             .syms = NULL,
+#endif
                              .screen_nbr = -1,
                              .sndisplay = NULL,
                              .sncontext = NULL,
-                             .monitors = NULL};
+                             .monitors = NULL,
+                             .clipboard = NULL};
 xcb_stuff *xcb = &xcb_int;
 
 /**
@@ -115,13 +128,13 @@ const struct {
 } cursor_names[] = {
     {"default", "left_ptr"}, {"pointer", "hand"}, {"text", "xterm"}};
 
-static xcb_visualtype_t *lookup_visual(xcb_screen_t *s, xcb_visualid_t visual) {
+static xcb_visualtype_t *lookup_visual(xcb_screen_t *s, xcb_visualid_t vis) {
   xcb_depth_iterator_t d;
   d = xcb_screen_allowed_depths_iterator(s);
   for (; d.rem; xcb_depth_next(&d)) {
     xcb_visualtype_iterator_t v = xcb_depth_visuals_iterator(d.data);
     for (; v.rem; xcb_visualtype_next(&v)) {
-      if (v.data->visual_id == visual) {
+      if (v.data->visual_id == vis) {
         return v.data;
       }
     }
@@ -423,6 +436,23 @@ static void x11_monitors_free(void) {
   }
 }
 
+/**
+ * Quick function that tries to fix the size (for dpi calculation)
+ * when monitor is rotate. This assumes the density is kinda equal in both X/Y
+ * direction.
+ */
+static void x11_workarea_fix_rotation(workarea *w) {
+  double ratio_res = w->w / (double)w->h;
+  double ratio_size = w->mw / (double)w->mh;
+
+  if ((ratio_res < 1.0 && ratio_size > 1.0) ||
+      (ratio_res > 1.0 && ratio_size < 1.0)) {
+    // Oposite ratios, swap them.
+    int nh = w->mw;
+    w->mw = w->mh;
+    w->mh = nh;
+  }
+}
 /**
  * Create monitor based on output id
  */
@@ -451,6 +481,7 @@ static workarea *x11_get_monitor_from_output(xcb_randr_output_t out) {
 
   retv->mw = op_reply->mm_width;
   retv->mh = op_reply->mm_height;
+  x11_workarea_fix_rotation(retv);
 
   char *tname = (char *)xcb_randr_get_output_info_name(op_reply);
   int tname_len = xcb_randr_get_output_info_name_length(op_reply);
@@ -501,6 +532,7 @@ x11_get_monitor_from_randr_monitor(xcb_randr_monitor_info_t *mon) {
   // Physical
   retv->mw = mon->width_in_millimeters;
   retv->mh = mon->height_in_millimeters;
+  x11_workarea_fix_rotation(retv);
 
   // Name
   retv->name =
@@ -528,7 +560,7 @@ static int x11_is_extension_present(const char *extension) {
   return present;
 }
 
-static void x11_build_monitor_layout_xinerama() {
+static void x11_build_monitor_layout_xinerama(void) {
   xcb_xinerama_query_screens_cookie_t screens_cookie =
       xcb_xinerama_query_screens_unchecked(xcb->connection);
 
@@ -559,7 +591,7 @@ static void x11_build_monitor_layout_xinerama() {
   free(screens_reply);
 }
 
-static void x11_build_monitor_layout() {
+static void x11_build_monitor_layout(void) {
   if (xcb->monitors) {
     return;
   }
@@ -744,6 +776,10 @@ static int monitor_get_dimension(int monitor_id, workarea *mon) {
 }
 // find the dimensions of the monitor displaying point x,y
 static void monitor_dimensions(int x, int y, workarea *mon) {
+  if (mon == NULL) {
+    g_error("%s: mon == NULL", __func__);
+    return;
+  }
   memset(mon, 0, sizeof(workarea));
   mon->w = xcb->screen->width_in_pixels;
   mon->h = xcb->screen->height_in_pixels;
@@ -781,7 +817,12 @@ static int pointer_get(xcb_window_t root, int *x, int *y) {
 
   return FALSE;
 }
+
 static int monitor_active_from_winid(xcb_drawable_t id, workarea *mon) {
+  if (mon == NULL) {
+    g_error("%s: mon == NULL", __func__);
+    return FALSE;
+  }
   xcb_window_t root = xcb->screen->root;
   xcb_get_geometry_cookie_t c = xcb_get_geometry(xcb->connection, id);
   xcb_get_geometry_reply_t *r =
@@ -811,6 +852,10 @@ static int monitor_active_from_id_focused(int mon_id, workarea *mon) {
   int retv = FALSE;
   xcb_window_t active_window;
   xcb_get_property_cookie_t awc;
+  if (mon == NULL) {
+    g_error("%s: mon == NULL", __func__);
+    return retv;
+  }
   awc = xcb_ewmh_get_active_window(&xcb->ewmh, xcb->screen_nbr);
   if (!xcb_ewmh_get_active_window_reply(&xcb->ewmh, awc, &active_window,
                                         NULL)) {
@@ -838,35 +883,38 @@ static int monitor_active_from_id_focused(int mon_id, workarea *mon) {
     free(tree_reply);
     return retv;
   }
-  xcb_translate_coordinates_cookie_t ct = xcb_translate_coordinates(
-      xcb->connection, tree_reply->parent, r->root, r->x, r->y);
-  xcb_translate_coordinates_reply_t *t =
-      xcb_translate_coordinates_reply(xcb->connection, ct, NULL);
-  if (t) {
-    if (mon_id == -2) {
-      // place the menu above the window
-      // if some window is focused, place menu above window, else fall
-      // back to selected monitor.
-      mon->x = t->dst_x - r->x;
-      mon->y = t->dst_y - r->y;
-      mon->w = r->width;
-      mon->h = r->height;
-      retv = TRUE;
-      if ((current_window_manager & WM_ROOT_WINDOW_OFFSET) ==
-          WM_ROOT_WINDOW_OFFSET) {
-        mon->x += r->x;
-        mon->y += r->y;
-      }
-      g_debug("mon pos: %d %d %d-%d", mon->x, mon->y, mon->w, mon->h);
-    } else if (mon_id == -4) {
-      monitor_dimensions(t->dst_x, t->dst_y, mon);
-      retv = TRUE;
+  if (tree_reply->parent != r->root) {
+    xcb_translate_coordinates_cookie_t ct = xcb_translate_coordinates(
+        xcb->connection, tree_reply->parent, r->root, r->x, r->y);
+    xcb_translate_coordinates_reply_t *t =
+        xcb_translate_coordinates_reply(xcb->connection, ct, NULL);
+    if (t) {
+      r->x = t->dst_x;
+      r->y = t->dst_y;
+      free(t);
+    } else {
+      g_debug("Failed to get translate position of active window, falling back "
+              "to mouse location (-5).");
+      free(r);
+      free(tree_reply);
+      return retv;
     }
-    free(t);
-  } else {
-    g_debug("Failed to get translate position of active window, falling back "
-            "to mouse location (-5).");
   }
+  if (mon_id == -2) {
+    // place the menu above the window
+    // if some window is focused, place menu above window, else fall
+    // back to selected monitor.
+    mon->x = r->x + r->border_width;
+    mon->y = r->y + r->border_width;
+    mon->w = r->width;
+    mon->h = r->height;
+    retv = TRUE;
+  } else if (mon_id == -4) {
+    g_debug("Find monitor at location: %d %d", r->x, r->y);
+    monitor_dimensions(r->x, r->y, mon);
+    retv = TRUE;
+  }
+  g_debug("mon pos: %d %d %d-%d", mon->x, mon->y, mon->w, mon->h);
   free(r);
   free(tree_reply);
   return retv;
@@ -874,6 +922,11 @@ static int monitor_active_from_id_focused(int mon_id, workarea *mon) {
 static int monitor_active_from_id(int mon_id, workarea *mon) {
   xcb_window_t root = xcb->screen->root;
   int x, y;
+  if (mon == NULL) {
+    g_error("%s: mon == NULL", __func__);
+    return FALSE;
+  }
+  g_debug("Monitor id: %d", mon_id);
   // At mouse position.
   if (mon_id == -3) {
     if (pointer_get(root, &x, &y)) {
@@ -885,19 +938,27 @@ static int monitor_active_from_id(int mon_id, workarea *mon) {
   }
   // Focused monitor
   else if (mon_id == -1) {
+    g_debug("rofi on current monitor");
     // Get the current desktop.
     unsigned int current_desktop = 0;
     xcb_get_property_cookie_t gcdc;
     gcdc = xcb_ewmh_get_current_desktop(&xcb->ewmh, xcb->screen_nbr);
     if (xcb_ewmh_get_current_desktop_reply(&xcb->ewmh, gcdc, &current_desktop,
                                            NULL)) {
+      g_debug("Found current desktop: %u", current_desktop);
       xcb_get_property_cookie_t c =
           xcb_ewmh_get_desktop_viewport(&xcb->ewmh, xcb->screen_nbr);
       xcb_ewmh_get_desktop_viewport_reply_t vp;
       if (xcb_ewmh_get_desktop_viewport_reply(&xcb->ewmh, c, &vp, NULL)) {
+        g_debug("Found %d number of desktops", vp.desktop_viewport_len);
         if (current_desktop < vp.desktop_viewport_len) {
+          g_debug("Found viewport for desktop: %d %d",
+                  vp.desktop_viewport[current_desktop].x,
+                  vp.desktop_viewport[current_desktop].y);
           monitor_dimensions(vp.desktop_viewport[current_desktop].x,
                              vp.desktop_viewport[current_desktop].y, mon);
+          g_debug("Found monitor @: %d %d %dx%d", mon->x, mon->y, mon->w,
+                  mon->h);
           xcb_ewmh_get_desktop_viewport_reply_wipe(&vp);
           return TRUE;
         }
@@ -934,20 +995,41 @@ static int monitor_active_from_id(int mon_id, workarea *mon) {
 
 // determine which monitor holds the active window, or failing that the mouse
 // pointer
+/** The cached monitor setup (mon_cache) is populated */
+gboolean mon_set = FALSE;
+/** cached monitor cache, to avoid multiple roundtrips to fetch this. */
+workarea mon_cache = {
+    0,
+};
 int monitor_active(workarea *mon) {
+  if (mon == NULL) {
+    g_error("%s: mon == NULL", __func__);
+    return FALSE;
+  }
+  g_debug("Monitor active");
+  if (mon_set) {
+    *mon = mon_cache;
+    return TRUE;
+  }
   if (config.monitor != NULL) {
+    g_debug("Monitor lookup  by name : %s", config.monitor);
     for (workarea *iter = xcb->monitors; iter; iter = iter->next) {
       if (g_strcmp0(config.monitor, iter->name) == 0) {
         *mon = *iter;
+        mon_cache = *mon;
+        mon_set = TRUE;
         return TRUE;
       }
     }
   }
+  g_debug("Monitor lookup  by name failed: %s", config.monitor);
   // Grab primary.
   if (g_strcmp0(config.monitor, "primary") == 0) {
     for (workarea *iter = xcb->monitors; iter; iter = iter->next) {
       if (iter->primary) {
         *mon = *iter;
+        mon_cache = *mon;
+        mon_set = TRUE;
         return TRUE;
       }
     }
@@ -957,6 +1039,8 @@ int monitor_active(workarea *mon) {
     xcb_drawable_t win = g_ascii_strtoll(config.monitor + 4, &end, 0);
     if (end != config.monitor) {
       if (monitor_active_from_winid(win, mon)) {
+        mon_cache = *mon;
+        mon_set = TRUE;
         return TRUE;
       }
     }
@@ -968,19 +1052,55 @@ int monitor_active(workarea *mon) {
     if (end != config.monitor) {
       if (mon_id >= 0) {
         if (monitor_get_dimension(mon_id, mon)) {
+          mon_cache = *mon;
+          mon_set = TRUE;
           return TRUE;
         }
         g_warning("Failed to find selected monitor.");
       } else {
-        return monitor_active_from_id(mon_id, mon);
+        int val = monitor_active_from_id(mon_id, mon);
+        mon_cache = *mon;
+        mon_set = TRUE;
+        return val;
       }
     }
   }
   // Fallback.
   monitor_dimensions(0, 0, mon);
+  mon_cache = *mon;
+  mon_set = TRUE;
   return FALSE;
 }
 
+static bool get_atom_name(xcb_connection_t *conn, xcb_atom_t atom, char **out) {
+  xcb_get_atom_name_cookie_t cookie;
+  xcb_get_atom_name_reply_t *reply;
+  int length;
+  char *name;
+
+  if (atom == 0) {
+    *out = NULL;
+    return true;
+  }
+
+  cookie = xcb_get_atom_name(conn, atom);
+  reply = xcb_get_atom_name_reply(conn, cookie, NULL);
+  if (!reply)
+    return false;
+
+  length = xcb_get_atom_name_name_length(reply);
+  name = xcb_get_atom_name_name(reply);
+
+  (*out) = g_strndup(name, length);
+  if (!(*out)) {
+    free(reply);
+    return false;
+  }
+
+  free(reply);
+  return true;
+}
+
 /**
  * @param state Internal state of the menu.
  * @param xse   X selection event.
@@ -990,7 +1110,7 @@ int monitor_active(workarea *mon) {
 static void rofi_view_paste(RofiViewState *state,
                             xcb_selection_notify_event_t *xse) {
   if (xse->property == XCB_ATOM_NONE) {
-    g_warning("Failed to convert selection");
+    g_debug("Failed to convert selection");
   } else if (xse->property == xcb->ewmh.UTF8_STRING) {
     gchar *text = window_get_text_prop(xse->requestor, xcb->ewmh.UTF8_STRING);
     if (text != NULL && text[0] != '\0') {
@@ -1005,7 +1125,13 @@ static void rofi_view_paste(RofiViewState *state,
     }
     g_free(text);
   } else {
-    g_warning("Failed");
+    char *out = NULL;
+    if (get_atom_name(xcb->connection, xse->property, &out)) {
+      g_debug("rofi_view_paste: Got unknown atom: %s", out);
+      g_free(out);
+    } else {
+      g_debug("rofi_view_paste: Got unknown, unnamed: %s", out);
+    }
   }
 }
 
@@ -1062,6 +1188,34 @@ static gboolean x11_button_to_nk_bindings_scroll(guint32 x11_button,
   return TRUE;
 }
 
+static void rofi_key_press_event_handler(xcb_key_press_event_t *xkpe,
+                                         RofiViewState *state) {
+  gchar *text;
+  g_log("IMDKit", G_LOG_LEVEL_DEBUG, "press handler %d", xkpe->detail);
+
+  xcb->last_timestamp = xkpe->time;
+  if (config.xserver_i300_workaround) {
+    text = nk_bindings_seat_handle_key_with_modmask(
+        xcb->bindings_seat, NULL, xkpe->state, xkpe->detail,
+        NK_BINDINGS_KEY_STATE_PRESS);
+  } else {
+    text = nk_bindings_seat_handle_key(xcb->bindings_seat, NULL, xkpe->detail,
+                                       NK_BINDINGS_KEY_STATE_PRESS);
+  }
+  if (text != NULL) {
+    rofi_view_handle_text(state, text);
+    g_free(text);
+  }
+}
+
+static void rofi_key_release_event_handler(xcb_key_release_event_t *xkre,
+                                           G_GNUC_UNUSED RofiViewState *state) {
+  g_log("IMDKit", G_LOG_LEVEL_DEBUG, "release handler %d", xkre->detail);
+  xcb->last_timestamp = xkre->time;
+  nk_bindings_seat_handle_key(xcb->bindings_seat, NULL, xkre->detail,
+                              NK_BINDINGS_KEY_STATE_RELEASE);
+}
+
 /**
  * Process X11 events in the main-loop (gui-thread) of the application.
  */
@@ -1083,6 +1237,26 @@ static void main_loop_x11_event_handler_view(xcb_generic_event_t *event) {
     }
     break;
   }
+  case XCB_DESTROY_NOTIFY: {
+    xcb_window_t win = ((xcb_destroy_notify_event_t *)event)->window;
+    if (win != rofi_view_get_window()) {
+#ifdef WINDOW_MODE
+      window_client_handle_signal(win, FALSE);
+#endif
+    } else {
+      g_main_loop_quit(xcb->main_loop);
+    }
+    break;
+  }
+  case XCB_CREATE_NOTIFY: {
+    xcb_window_t win = ((xcb_create_notify_event_t *)event)->window;
+    if (win != rofi_view_get_window()) {
+#ifdef WINDOW_MODE
+      window_client_handle_signal(win, TRUE);
+#endif
+    }
+    break;
+  }
   case XCB_EXPOSE:
     rofi_view_frame_callback();
     break;
@@ -1094,9 +1268,6 @@ static void main_loop_x11_event_handler_view(xcb_generic_event_t *event) {
   case XCB_MOTION_NOTIFY: {
     xcb_motion_notify_event_t *xme = (xcb_motion_notify_event_t *)event;
     gboolean button_mask = xme->state & XCB_EVENT_MASK_BUTTON_1_MOTION;
-    if (button_mask && config.click_to_exit == TRUE) {
-      xcb->mouse_seen = TRUE;
-    }
     rofi_view_handle_mouse_motion(state, xme->event_x, xme->event_y,
                                   !button_mask && config.hover_select);
     break;
@@ -1115,8 +1286,55 @@ static void main_loop_x11_event_handler_view(xcb_generic_event_t *event) {
     } else if (x11_button_to_nk_bindings_scroll(bpe->detail, &axis, &steps)) {
       nk_bindings_seat_handle_scroll(xcb->bindings_seat, NULL, axis, steps);
     }
+    xcb->mouse_seen++;
     break;
   }
+  case XCB_SELECTION_CLEAR: {
+    g_debug("Selection Clear.");
+    xcb_stuff_set_clipboard(NULL);
+  } break;
+  case XCB_SELECTION_REQUEST: {
+    g_debug("Selection Request.");
+    xcb_selection_request_event_t *req = (xcb_selection_request_event_t *)event;
+    if (req->selection == netatoms[CLIPBOARD]) {
+      xcb_atom_t targets[2];
+      xcb_selection_notify_event_t selection_notify = {
+          .response_type = XCB_SELECTION_NOTIFY,
+          .sequence = 0,
+          .time = req->time,
+          .requestor = req->requestor,
+          .selection = req->selection,
+          .target = req->target,
+          .property = XCB_ATOM_NONE,
+      };
+      // If no clipboard, we return NONE.
+      if (xcb->clipboard) {
+        // Request for UTF-8
+        if (req->target == netatoms[UTF8_STRING]) {
+          g_debug("Selection Request UTF-8.");
+          xcb_change_property(xcb->connection, XCB_PROP_MODE_REPLACE,
+                              req->requestor, req->property,
+                              netatoms[UTF8_STRING], 8,
+                              strlen(xcb->clipboard) + 1, xcb->clipboard);
+          selection_notify.property = req->property;
+        } else if (req->target == netatoms[TARGETS]) {
+          g_debug("Selection Request Targets.");
+          // We currently only support UTF8 from clipboard. So indicate this.
+          targets[0] = netatoms[UTF8_STRING];
+          xcb_change_property(xcb->connection, XCB_PROP_MODE_REPLACE,
+                              req->requestor, req->property, XCB_ATOM_ATOM, 32,
+                              1, targets);
+          selection_notify.property = req->property;
+        }
+      }
+
+      xcb_send_event(xcb->connection,
+                     0, // propagate
+                     req->requestor, XCB_EVENT_MASK_NO_EVENT,
+                     (const char *)&selection_notify);
+      xcb_flush(xcb->connection);
+    }
+  } break;
   case XCB_BUTTON_RELEASE: {
     xcb_button_release_event_t *bre = (xcb_button_release_event_t *)event;
     NkBindingsMouseButton button;
@@ -1131,7 +1349,7 @@ static void main_loop_x11_event_handler_view(xcb_generic_event_t *event) {
       if (!xcb->mouse_seen) {
         rofi_view_temp_click_to_exit(state, bre->event);
       }
-      xcb->mouse_seen = FALSE;
+      xcb->mouse_seen--;
     }
     break;
   }
@@ -1155,22 +1373,40 @@ static void main_loop_x11_event_handler_view(xcb_generic_event_t *event) {
   }
   case XCB_KEY_PRESS: {
     xcb_key_press_event_t *xkpe = (xcb_key_press_event_t *)event;
-    gchar *text;
-
-    xcb->last_timestamp = xkpe->time;
-    text = nk_bindings_seat_handle_key_with_modmask(
-        xcb->bindings_seat, NULL, xkpe->state, xkpe->detail,
-        NK_BINDINGS_KEY_STATE_PRESS);
-    if (text != NULL) {
-      rofi_view_handle_text(state, text);
+#ifdef XCB_IMDKIT
+    if (xcb->ic) {
+      g_log("IMDKit", G_LOG_LEVEL_DEBUG, "press key %d to xim", xkpe->detail);
+      xcb_xim_forward_event(xcb->im, xcb->ic, xkpe);
+      return;
+    } else
+#endif
+    {
+      rofi_key_press_event_handler(xkpe, state);
     }
     break;
   }
   case XCB_KEY_RELEASE: {
     xcb_key_release_event_t *xkre = (xcb_key_release_event_t *)event;
-    xcb->last_timestamp = xkre->time;
-    nk_bindings_seat_handle_key(xcb->bindings_seat, NULL, xkre->detail,
-                                NK_BINDINGS_KEY_STATE_RELEASE);
+#ifdef XCB_IMDKIT
+    if (xcb->ic) {
+      g_log("IMDKit", G_LOG_LEVEL_DEBUG, "release key %d to xim", xkre->detail);
+
+      // Check if the keysym is a modifier key (e.g., Shift, Ctrl, Alt). If it
+      // is, sleep for 5 milliseconds as a workaround for XCB XIM limitation.
+      // This sleep helps to ensure that XCB XIM can properly handle subsequent
+      // key events that may occur rapidly after a modifier key is pressed.
+      xcb_keysym_t sym = xcb_key_press_lookup_keysym(xcb->syms, xkre, 0);
+      if (xcb_is_modifier_key(sym)) {
+        struct timespec five_millis = {.tv_sec = 0, .tv_nsec = 5000000};
+        nanosleep(&five_millis, NULL);
+      }
+      xcb_xim_forward_event(xcb->im, xcb->ic, xkre);
+      return;
+    } else
+#endif
+    {
+      rofi_key_release_event_handler(xkre, state);
+    }
     break;
   }
   default:
@@ -1179,6 +1415,27 @@ static void main_loop_x11_event_handler_view(xcb_generic_event_t *event) {
   rofi_view_maybe_update(state);
 }
 
+#ifdef XCB_IMDKIT
+void x11_event_handler_fowarding(G_GNUC_UNUSED xcb_xim_t *im,
+                                 G_GNUC_UNUSED xcb_xic_t ic,
+                                 xcb_key_press_event_t *event,
+                                 G_GNUC_UNUSED void *user_data) {
+  RofiViewState *state = rofi_view_get_active();
+  if (state == NULL) {
+    return;
+  }
+
+  uint8_t type = event->response_type & ~0x80;
+  if (type == XCB_KEY_PRESS) {
+    rofi_key_press_event_handler(event, state);
+  } else if (type == XCB_KEY_RELEASE) {
+    xcb_key_release_event_t *xkre = (xcb_key_release_event_t *)event;
+    rofi_key_release_event_handler(xkre, state);
+  }
+  rofi_view_maybe_update(state);
+}
+#endif
+
 static gboolean main_loop_x11_event_handler(xcb_generic_event_t *ev,
                                             G_GNUC_UNUSED gpointer user_data) {
   if (ev == NULL) {
@@ -1188,9 +1445,18 @@ static gboolean main_loop_x11_event_handler(xcb_generic_event_t *ev,
       g_main_loop_quit(xcb->main_loop);
       return G_SOURCE_REMOVE;
     }
-    g_warning("main_loop_x11_event_handler: ev == NULL, status == %d", status);
+    // DD: it seems this handler often gets dispatched while the queue in GWater
+    // is empty. resulting in a NULL for ev. This seems not an error.
+    // g_warning("main_loop_x11_event_handler: ev == NULL, status == %d",
+    // status);
     return G_SOURCE_CONTINUE;
   }
+
+#ifdef XCB_IMDKIT
+  if (xcb->im && xcb_xim_filter_event(xcb->im, ev))
+    return G_SOURCE_CONTINUE;
+#endif
+
   uint8_t type = ev->response_type & ~0x80;
   if (type == xcb->xkb.first_event) {
     switch (ev->pad0) {
@@ -1220,6 +1486,7 @@ static gboolean main_loop_x11_event_handler(xcb_generic_event_t *ev,
   if (xcb->sndisplay != NULL) {
     sn_xcb_display_process_event(xcb->sndisplay, ev);
   }
+
   main_loop_x11_event_handler_view(ev);
   return G_SOURCE_CONTINUE;
 }
@@ -1356,6 +1623,26 @@ static void x11_create_frequently_used_atoms(void) {
   }
 }
 
+char *x11_helper_get_window_manager(void) {
+  char *retv = NULL;
+  xcb_window_t wm_win = 0;
+  xcb_get_property_cookie_t cc = xcb_ewmh_get_supporting_wm_check_unchecked(
+      &xcb->ewmh, xcb_stuff_get_root_window());
+
+  if (xcb_ewmh_get_supporting_wm_check_reply(&xcb->ewmh, cc, &wm_win, NULL)) {
+    xcb_ewmh_get_utf8_strings_reply_t wtitle;
+    xcb_get_property_cookie_t cookie =
+        xcb_ewmh_get_wm_name_unchecked(&(xcb->ewmh), wm_win);
+    if (xcb_ewmh_get_wm_name_reply(&(xcb->ewmh), cookie, &wtitle, (void *)0)) {
+      if (wtitle.strings_len > 0) {
+        retv = g_strndup(wtitle.strings, wtitle.strings_len);
+      }
+      xcb_ewmh_get_utf8_strings_reply_wipe(&wtitle);
+    }
+  }
+  return retv;
+}
+
 static void x11_helper_discover_window_manager(void) {
   xcb_window_t wm_win = 0;
   xcb_get_property_cookie_t cc = xcb_ewmh_get_supporting_wm_check_unchecked(
@@ -1367,13 +1654,14 @@ static void x11_helper_discover_window_manager(void) {
         xcb_ewmh_get_wm_name_unchecked(&(xcb->ewmh), wm_win);
     if (xcb_ewmh_get_wm_name_reply(&(xcb->ewmh), cookie, &wtitle, (void *)0)) {
       if (wtitle.strings_len > 0) {
-        g_debug("Found window manager: |%s|", wtitle.strings);
-        if (g_strcmp0(wtitle.strings, "i3") == 0) {
+        // Copy the string and add terminating '\0'.
+        char *str = g_strndup(wtitle.strings, wtitle.strings_len);
+        g_debug("Found window manager: |%s|", str);
+        if (g_strcmp0(str, "i3") == 0) {
           current_window_manager =
               WM_DO_NOT_CHANGE_CURRENT_DESKTOP | WM_PANGO_WORKSPACE_NAMES;
-        } else if (g_strcmp0(wtitle.strings, "bspwm") == 0) {
-          current_window_manager = WM_ROOT_WINDOW_OFFSET;
         }
+        g_free(str);
       }
       xcb_ewmh_get_utf8_strings_reply_wipe(&wtitle);
     }
@@ -1387,6 +1675,9 @@ gboolean display_setup(GMainLoop *main_loop, NkBindings *bindings) {
   find_arg_str("-display", &display_str);
 
   xcb->main_loop = main_loop;
+#ifdef XCB_IMDKIT
+  xcb_compound_text_init();
+#endif
   xcb->source = g_water_xcb_source_new(g_main_loop_get_context(xcb->main_loop),
                                        display_str, &xcb->screen_nbr,
                                        main_loop_x11_event_handler, NULL, NULL);
@@ -1395,6 +1686,17 @@ gboolean display_setup(GMainLoop *main_loop, NkBindings *bindings) {
     return FALSE;
   }
   xcb->connection = g_water_xcb_source_get_connection(xcb->source);
+#ifdef XCB_IMDKIT
+  xcb->im = xcb_xim_create(xcb->connection, xcb->screen_nbr, NULL);
+  xcb->syms = xcb_key_symbols_alloc(xcb->connection);
+#endif
+
+#ifdef XCB_IMDKIT
+#ifndef XCB_IMDKIT_1_0_3_LOWER
+  xcb_xim_set_use_compound_text(xcb->im, true);
+  xcb_xim_set_use_utf8_string(xcb->im, true);
+#endif
+#endif
 
   TICK_N("Open Display");
 
@@ -1484,6 +1786,11 @@ gboolean display_setup(GMainLoop *main_loop, NkBindings *bindings) {
     return FALSE;
   }
 
+  uint32_t val[] = {XCB_EVENT_MASK_SUBSTRUCTURE_NOTIFY};
+
+  xcb_change_window_attributes(xcb->connection, xcb_stuff_get_root_window(),
+                               XCB_CW_EVENT_MASK, val);
+
   // startup not.
   xcb->sndisplay =
       sn_xcb_display_new(xcb->connection, error_trap_push, error_trap_pop);
@@ -1609,7 +1916,7 @@ gboolean display_late_setup(void) {
   // Try to grab the keyboard as early as possible.
   // We grab this using the rootwindow (as dmenu does it).
   // this seems to result in the smallest delay for most people.
-  if (find_arg("-normal-window") >= 0) {
+  if (find_arg("-normal-window") >= 0 || find_arg("-transient-window") >= 0) {
     return TRUE;
   }
   if (find_arg("-no-lazy-grab") >= 0) {
@@ -1659,6 +1966,11 @@ void display_cleanup(void) {
   xcb_ewmh_connection_wipe(&(xcb->ewmh));
   xcb_flush(xcb->connection);
   xcb_aux_sync(xcb->connection);
+#ifdef XCB_IMDKIT
+  xcb_xim_close(xcb->im);
+  xcb_xim_destroy(xcb->im);
+  xcb->im = NULL;
+#endif
   g_water_xcb_source_free(xcb->source);
   xcb->source = NULL;
   xcb->connection = NULL;
@@ -1702,3 +2014,7 @@ void x11_set_cursor(xcb_window_t window, X11CursorType type) {
   xcb_change_window_attributes(xcb->connection, window, XCB_CW_CURSOR,
                                &(cursors[type]));
 }
+void xcb_stuff_set_clipboard(char *data) {
+  g_free(xcb->clipboard);
+  xcb->clipboard = data;
+}
